import asyncio

from langchain.llms.base import BaseLLM
from langchain.schema import LLMResult

from langchain_visualizer.hijacking import ice_hijack


def overridden_generate(
    self, prompts: list[str], stop: list[str] | None = None
) -> LLMResult:
    """Preserve langchain's sync generate method"""
    ice_agent = self.get_ice_agent()
    generations = []
    llm_output = {}
    for prompt in prompts:
        llm_result = asyncio.get_event_loop().run_until_complete(
            ice_agent.generate(prompt=prompt, stop=stop)
        )
        generations.extend(llm_result.generations)
        llm_output.update(llm_result.llm_output)
    return LLMResult(generations, llm_output)


ice_hijack(BaseLLM, "generate")



import asyncio

import gorilla
from ice.trace import TracedABC
from pydantic import Extra

LANGCHAIN_VISUALIZER_PATCH_ID = "lc-viz"
VCR_LANGCHAIN_PATCH_ID = "lc-vcr"
# override prefix used by vcr-langchain for visualization compatibility
VCR_VIZ_INTEROP_PREFIX = "_vcr_"


class VisualizationWrapper(TracedABC):
    def __init__(self, og_obj, og_fn):
        self.og_obj = og_obj
        self.og_fn = og_fn

    async def run(self, *args, **kwargs):
        return self.og_fn(self.og_obj, *args, **kwargs)


class PydanticBypass:
    def __init__(self, obj):
        self.obj = obj
        self.has_config = hasattr(self.obj.__class__, "__config__")
        self.has_extra = self.has_config and hasattr(
            self.obj.__class__.__config__, "extra"
        )

    def __enter__(self):
        if self.has_extra:
            self.obj.__class__.__config__.extra = Extra.allow

    def __exit__(self, *args):
        if self.has_extra:
            self.obj.__class__.__config__.extra = Extra.forbid


def ice_agent_getter(og_fn_name: str):
    def get_ice_agent(self):
        # we have to do this roundabout thing because we don't know ahead of time all
        # the classes we want to set ice_agent for
        if not hasattr(self, "ice_agent"):
            with PydanticBypass(self):
                vcr_key = VCR_VIZ_INTEROP_PREFIX + og_fn_name
                if hasattr(self.__class__, vcr_key):
                    # if vcr-langchain is here as well, call them so that caching can
                    # happen
                    og_fn = getattr(self.__class__, vcr_key)
                else:
                    og_fn = gorilla.get_original_attribute(
                        self.__class__, og_fn_name, LANGCHAIN_VISUALIZER_PATCH_ID
                    )
                self.ice_agent = VisualizationWrapper(self, og_fn=og_fn)
        return self.ice_agent

    return get_ice_agent


def overridden_call(self, *args, **kwargs):
    """Preserve sync nature of OG call method"""
    ice_agent = self.get_ice_agent()
    if not hasattr(self.__class__, "_should_trace") or self.__class__._should_trace:
        # ICE displays class name in visualization
        ice_agent.__class__.__name__ = self.__class__.__name__
        return asyncio.get_event_loop().run_until_complete(
            ice_agent.run(*args, **kwargs)
        )

    return ice_agent.og_fn(self, *args, **kwargs)


def hijack(cls, fn_name, get_replacement):
    replacement = get_replacement(getattr(cls, fn_name))
    setattr(cls, fn_name, replacement)


def ice_hijack(cls, og_method_name):
    gorilla.apply(
        gorilla.Patch(
            destination=cls,
            name="get_ice_agent",
            obj=ice_agent_getter(og_method_name),
            settings=gorilla.Settings(allow_hit=False),
        )
    )
    gorilla.apply(
        gorilla.Patch(
            destination=cls,
            name=og_method_name,
            obj=overridden_call,
            settings=gorilla.Settings(store_hit=True, allow_hit=True),
        ),
        id=LANGCHAIN_VISUALIZER_PATCH_ID,
    )



from fvalues import F, FValue
from langchain import PromptTemplate
from langchain.formatting import formatter

from langchain_visualizer.hijacking import hijack


def format_f(formatter, string, *args, **kwargs) -> F:
    # if there are any issues with the formatting, let the formatter expose them first
    result = formatter.format(string, *args, **kwargs)
    parts = []
    # modified from string._vformat
    for literal_text, field_name, format_spec, _ in formatter.parse(string):
        if literal_text:
            parts.append(literal_text)

        if field_name is not None:
            obj, _ = formatter.get_field(field_name, args, kwargs)
            parts.append(
                FValue(
                    source=field_name,
                    value=obj,
                    formatted=formatter.format_field(obj, format_spec),
                )
            )

    return F(result, parts=tuple(parts))


def get_new_format(og_format):
    def new_format(self, *args, **kwargs) -> str:
        if self.template_format != "f-string":
            return og_format(*args, **kwargs)

        return format_f(formatter, self.template, *args, **kwargs)

    return new_format


hijack(PromptTemplate, "format", get_new_format)



from langchain.python import PythonREPL
from langchain.serpapi import SerpAPIWrapper

from langchain_visualizer.hijacking import ice_hijack

ice_hijack(SerpAPIWrapper, "run")
ice_hijack(PythonREPL, "run")



import string
from typing import Optional

from fvalues import F, FValue
from langchain import FewShotPromptTemplate

from langchain_visualizer.hijacking import hijack
from langchain_visualizer.prompts.prompt import format_f


class FBuilder:
    def __init__(self, segments: Optional[list[F | str]] = None):
        self.f_values = segments or []

    def add_segment(self, segment: F | str):
        if not hasattr(segment, "parts"):
            segment = F(segment, parts=(segment,))
        self.f_values.append(segment)

    def pop(self):
        return self.f_values.pop()

    def build(self):
        finaL_str = "".join(self.f_values)
        final_parts = []
        previous_was_str = False
        for f_value in self.f_values:
            for part in f_value.flatten().parts:
                if isinstance(part, FValue):
                    final_parts.append(part)
                    previous_was_str = False
                else:  # it's a simple str
                    if previous_was_str:  # concatenate the two strings
                        final_parts[-1] = final_parts[-1] + part
                    else:
                        final_parts.append(part)

                    previous_was_str = True

        return F(finaL_str, parts=tuple(final_parts))


def get_new_format(og_format):
    def new_format(self, *args, **kwargs) -> str:
        if self.template_format != "f-string":
            return og_format(*args, **kwargs)

        f_builder = FBuilder()
        if self.prefix:
            f_builder.add_segment(
                format_f(string.Formatter(), self.prefix, *args, **kwargs)
            )
            f_builder.add_segment(self.example_separator)

        examples = self._get_examples(**kwargs)
        # ignore nested templating for now
        for example in examples:
            example_f = self.example_prompt.format(**example)
            f_builder.add_segment(example_f)
            f_builder.add_segment(self.example_separator)
        f_builder.pop()  # pop last separator in case there's no suffix

        if self.suffix:
            f_builder.add_segment(self.example_separator)
            f_builder.add_segment(
                format_f(string.Formatter(), self.suffix, *args, **kwargs)
            )

        return f_builder.build()

    return new_format


hijack(FewShotPromptTemplate, "format", get_new_format)



from typing import Any, List

from ice import json_value
from langchain.schema import LLMResult

og_json_value = json_value.to_json_value


def to_json_value(x: Any) -> json_value.JSONValue:
    if isinstance(x, LLMResult):
        generations = x.generations
        texts: List[List[str]] | List[str] | str = [
            g.text for sublist in generations for g in sublist
        ]
        if len(texts) == 1:
            texts = texts[0]
        if len(texts) == 1:
            # do it a second time because it's a list of lists
            texts = texts[0]
        return og_json_value(texts)

    return og_json_value(x)


json_value.to_json_value = to_json_value

# this import cannot happen earlier because otherwise ice.json_value.to_json_value
# does not get successfully overwritten
from ice.recipe import recipe  # noqa: E402

visualize = recipe.main



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['langchain_visualizer',
 'langchain_visualizer.agents',
 'langchain_visualizer.chains',
 'langchain_visualizer.llms',
 'langchain_visualizer.prompts']

package_data = \
{'': ['*']}

install_requires = \
['gorilla>=0.4.0,<0.5.0',
 'langchain>=0.0.66,<0.0.67',
 'ought-ice>=0.4.0,<0.5.0']

setup_kwargs = {
    'name': 'langchain-visualizer',
    'version': '0.0.3',
    'description': 'Visualization and debugging tool for LangChain workflows',
    'long_description': '# LangChain Visualizer\n\nAdapts [ICE](https://github.com/oughtinc/ice) for use with [LangChain](https://github.com/hwchase17/langchain) so that you can view LangChain interactions with a beautiful UI.\n\n![Screenshot of an execution run](screenshots/serp_screenshot.png "SERP agent demonstration")\n\nYou can now\n\n- See the full prompt text being sent with every interaction with the LLM\n- Tell from the coloring which parts of the prompt are hardcoded and which parts are templated substitutions\n- Inspect the execution flow and observe when each function goes up the stack\n\n## Quickstart\n\nInstall this library:\n\n```bash\npip install langchain-visualizer\n```\n\nThen:\n\n1. Add `import langchain_visualizer` as **the first import** in your Python entrypoint file\n2. Write an async function to visualize whichever workflow you\'re running\n3. Call `langchain_visualizer.visualize` on that function\n\nFor an example, see below instructions on reproducing the screenshot.\n\n\n### Running the example screenshot\n\nTo run the example you see in the screenshot, first install this library and optional dependencies:\n\n```bash\npip install langchain-visualizer google-search-results openai\n```\n\nIf you haven\'t yet set up your [OpenAI API keys](https://openai.com/api/) or SERP API keys, you can [replay the recorded interactions](https://github.com/amosjyng/vcr-langchain) by cloning this repository and running\n\n```bash\npython tests/agents/test_langchain_getting_started.py\n```\n\nIf you have set them up, you can run the following script (adapted from [LangChain docs](https://langchain.readthedocs.io/en/latest/modules/agents/getting_started.html)):\n\n```python\nimport langchain_visualizer\nimport asyncio\nfrom langchain.agents import initialize_agent, load_tools\nfrom langchain.llms import OpenAI\n\nllm = OpenAI(temperature=0.7)\ntools = load_tools(["serpapi", "llm-math"], llm=llm)\nagent = initialize_agent(tools, llm, agent="zero-shot-react-description", verbose=True)\nasync def search_agent_demo():\n    return agent.run(\n        "Who is Olivia Wilde\'s boyfriend? What is his current age raised to the 0.23 "\n        "power?"\n    )\n\nlangchain_visualizer.visualize(search_agent_demo)\n```\n\nA browser window will open up, and you can actually see the agent execute happen in real-time!\n\n**Please note that there is a lot of langchain functionality that I haven\'t gotten around to hijacking for visualization.** If there\'s anything you need to show up in the execution trace, please open a PR or issue.\n\n## My other projects\n\nPlease check out [VCR LangChain](https://github.com/amosjyng/vcr-langchain), a library that lets you record LLM interactions for your tests and demos!\n',
    'author': 'Amos Jun-yeung Ng',
    'author_email': 'me@amos.ng',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.10,<3.11',
}


setup(**setup_kwargs)



# override ICE to_json_value before anything else starts importing other ICE stuff
# isort: off
from .ice import visualize as ice_visualize

# isort: on

from .agents.tools import SerpAPIWrapper  # noqa
from .chains.base import Chain  # noqa
from .llms.base import BaseLLM  # noqa
from .prompts.few_shot import FewShotPromptTemplate  # noqa
from .prompts.prompt import PromptTemplate  # noqa

visualize = ice_visualize
